/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2003 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Evan Nemerson <evan@coeus-group.com>                         |
  +----------------------------------------------------------------------+
*/

#define PHP_PANDA_EXT_VERSION "0.0.1"

/* $Id$ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_panda.h"

static int le_panda;
#define le_panda_name "Panda PDF"

static zend_class_entry *php_panda_ce_ptr;

/* {{{ Internal functions and macros
 *
 * These should only be used in this file.
 */
#define PHP_2_PANDA_BOOL(arg) (arg ? panda_true : panda_false)
#define PANDA_PHP_FETCH_PDF(PDF, pdf_arg) ZEND_FETCH_RESOURCE(PDF, php_panda_pdf *, &pdf_arg, -1, le_panda_name, le_panda)
#define php_panda_pdf_get_num_pages(x) x->pages

// Thanks Tidy (John Coggeshall)!
#define PANDA_METHOD_MAP(name, func_name, arg_types) \
	ZEND_NAMED_FE(name, ZEND_FN(func_name), arg_types)

static void destroy_panda_pdf(zend_rsrc_list_entry *rsrc TSRMLS_DC);
void panda_error(int fatal, char *msg);
static void php_panda_pdf_ff(php_panda_pdf *PDF);
static void php_panda_pdf_rw(php_panda_pdf *PDF);

static php_panda_pdf * php_panda_get_pdf(zval *obj TSRMLS_DC) {
	zval **tmp;
	int id_to_find;
	void *property;
	int type;

	if (obj) {
		if ( Z_TYPE_P(obj) != IS_OBJECT ) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid object.");
			return NULL;
		}
		if (zend_hash_find(Z_OBJPROP_P(obj), "pdf", 4, (void **)&tmp) == FAILURE) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to find property.");
			return NULL;
		}
		id_to_find = Z_LVAL_PP(tmp);
	} else {
		return NULL;
	}

	property = zend_list_find(id_to_find, &type);

	if (!property || type != le_panda) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to find identifier.");
		return NULL;
	}

	return property;
}

static void destroy_panda_pdf(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	php_panda_pdf *PDF;
	PDF = (rsrc->ptr);

	php_panda_pdf_rw(PDF);
	if ( PDF->pages > 0 ) {
		while ( PDF->currPage->nextPage != NULL ) {
			PDF->currPage = PDF->currPage->nextPage;
			efree(PDF->currPage->prevPage);
		}
		efree(PDF->currPage);
	}
	if ( PDF->PDF != NULL )
		panda_close(PDF->PDF);
	efree(PDF);
}

void panda_error(int fatal, char *message) {
	php_error_docref(NULL TSRMLS_CC, E_WARNING, "%s", message);
}

static void php_panda_pdf_ff(php_panda_pdf *PDF) {
	if ( PDF->pages > 0 )
		while ( PDF->currPage->nextPage != NULL ) 
			PDF->currPage = PDF->currPage->nextPage;
}

static void php_panda_pdf_rw(php_panda_pdf *PDF) {
	if ( PDF->pages > 0 )
		while ( PDF->currPage->prevPage != NULL )
			PDF->currPage = PDF->currPage->prevPage;
}
/* }}} */

/* {{{ panda_functions[]
 *
 * Every user visible function must have an entry in panda_functions[].
 */
function_entry panda_functions[] = {
	PHP_FE(panda_add_cubic_curve_segment,		NULL)
	PHP_FE(panda_add_line_segment,				NULL)
	PHP_FE(panda_add_quadratic_curve_segment,	NULL)
	PHP_FE(panda_center_window,					NULL)
	PHP_FE(panda_close,							NULL)
	PHP_FE(panda_close_line,					NULL)
	PHP_FE(panda_end_line,						NULL)
	PHP_FE(panda_fast_forward,					NULL)
	PHP_FE(panda_fill_line,						NULL)
	PHP_FE(panda_fit_window,					NULL)
	PHP_FE(panda_full_screen,					NULL)
	PHP_FE(panda_open,							NULL)
	PHP_FE(panda_get_num_pages,					NULL)
	PHP_FE(panda_get_page_height,				NULL)
	PHP_FE(panda_get_page_width,				NULL)
	PHP_FE(panda_hide_menu_bar,					NULL)
	PHP_FE(panda_hide_tool_bar,					NULL)
	PHP_FE(panda_hide_window_ui,				NULL)
	PHP_FE(panda_image_box,						NULL)
	PHP_FE(panda_link,			    			NULL)
	PHP_FE(panda_make_line,						NULL)
	PHP_FE(panda_make_page_size,				NULL)
	PHP_FE(panda_new_page,						NULL)
	PHP_FE(panda_next_page,						NULL)
	PHP_FE(panda_nfs_page_mode,					NULL)
	PHP_FE(panda_page_duration,					NULL)
	PHP_FE(panda_prev_page,						NULL)
	PHP_FE(panda_rectangle,						NULL)
	PHP_FE(panda_rewind,						NULL)
	PHP_FE(panda_set_author,					NULL)
	PHP_FE(panda_set_compression,				NULL)
	PHP_FE(panda_set_fill_color,				NULL)
	PHP_FE(panda_set_font,						NULL)
	PHP_FE(panda_set_font_size,					NULL)
	PHP_FE(panda_set_horizontal_scaling,		NULL)
	PHP_FE(panda_set_keywords,					NULL)
	PHP_FE(panda_set_leading,					NULL)
	PHP_FE(panda_set_line_cap,					NULL)
	PHP_FE(panda_set_line_color,				NULL)
	PHP_FE(panda_set_line_dash,					NULL)
	PHP_FE(panda_set_line_join,					NULL)
	PHP_FE(panda_set_line_start,				NULL)
	PHP_FE(panda_set_line_width,				NULL)
	PHP_FE(panda_set_subject,					NULL)
	PHP_FE(panda_set_text_mode,					NULL)
	PHP_FE(panda_set_title,						NULL)
	PHP_FE(panda_set_word_spacing,				NULL)
	PHP_FE(panda_stroke_line,					NULL)
	PHP_FE(panda_text_annotation,				NULL)
	PHP_FE(panda_text_box,						NULL)
	PHP_FE(panda_text_direction,				NULL)
	PHP_FE(panda_trans_duration,				NULL)
	PHP_FE(panda_trans_style,					NULL)
	PHP_FE(panda_uri_link,						NULL)
	/* The function in panda is panda_set_font_mode, but the constants are
	   panda_textmode_*. I wanted to be consistent with Michael's API, but
	   be forgiving as well... */
	PHP_FALIAS(panda_set_font_mode, panda_set_text_mode, NULL)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ panda_oo_functions[]
 *
 * Every user visible function must have an entry in panda_oo_functions[].
 */
#ifdef PHP_PANDA_OO
function_entry panda_oo_functions[] = {
    // Special methods.
	PANDA_METHOD_MAP(__construct,               panda_open,                        NULL)
	PHP_ME   (panda, __set,                     NULL,                              0)

	PANDA_METHOD_MAP(addCubicCurveSegment,      panda_add_cubic_curve_segment,     NULL)
	PANDA_METHOD_MAP(addLineSegment,            panda_add_line_segment,            NULL)
	PANDA_METHOD_MAP(addQuadraticCurveSegment,  panda_add_quadratic_curve_segment, NULL)
	PANDA_METHOD_MAP(centerWindow,              panda_center_window,               NULL)
	PANDA_METHOD_MAP(close,                     panda_close,                       NULL)
	PANDA_METHOD_MAP(closeLine,                 panda_close_line,                  NULL)
	PANDA_METHOD_MAP(endLine,                   panda_end_line,                    NULL)
	PANDA_METHOD_MAP(fastForward,               panda_fast_forward,                NULL)
	PANDA_METHOD_MAP(fillLine,                  panda_fill_line,                   NULL)
	PANDA_METHOD_MAP(fitWindow,                 panda_fit_window,                  NULL)
	PANDA_METHOD_MAP(fullScreen,                panda_full_screen,                 NULL)
	PANDA_METHOD_MAP(getNumPages,               panda_get_num_pages,               NULL)
	PANDA_METHOD_MAP(getPageHeight,             panda_get_page_height,             NULL)
	PANDA_METHOD_MAP(getPageWidth,              panda_get_page_width,              NULL)
	PANDA_METHOD_MAP(hideMenuBar,               panda_hide_menu_bar,               NULL)
	PANDA_METHOD_MAP(hideToolBar,               panda_hide_tool_bar,               NULL)
	PANDA_METHOD_MAP(hideWindowUI,              panda_hide_window_ui,              NULL)
	PANDA_METHOD_MAP(imageBox,                  panda_image_box,                   NULL)
	PANDA_METHOD_MAP(link,                      panda_link,                        NULL)
	PANDA_METHOD_MAP(makeLine,                  panda_make_line,                   NULL)
	PANDA_METHOD_MAP(makePageSize,              panda_make_page_size,              NULL)
	PANDA_METHOD_MAP(newPage,                   panda_new_page,                    NULL)
	PANDA_METHOD_MAP(nextPage,                  panda_next_page,                   NULL)
	PANDA_METHOD_MAP(nfsPageMode,               panda_nfs_page_mode,               NULL)
	PANDA_METHOD_MAP(pageDuration,              panda_page_duration,               NULL)
	PANDA_METHOD_MAP(prevPage,                  panda_prev_page,                   NULL)
	PANDA_METHOD_MAP(rectangle,                 panda_rectangle,                   NULL)
	PANDA_METHOD_MAP(rewind,                    panda_rewind,                      NULL)
	PANDA_METHOD_MAP(setAuthor,                 panda_set_author,                  NULL)
	PANDA_METHOD_MAP(setCompression,            panda_set_compression,             NULL)
	PANDA_METHOD_MAP(setFillColor,              panda_set_fill_color,              NULL)
	PANDA_METHOD_MAP(setFont,                   panda_set_font,                    NULL)
	PANDA_METHOD_MAP(setFontMode,               panda_set_text_mode,               NULL)
	PANDA_METHOD_MAP(setFontSize,               panda_set_font_size,               NULL)
	PANDA_METHOD_MAP(setHorizontalScaling,      panda_set_horizontal_scaling,      NULL)
	PANDA_METHOD_MAP(setKeywords,               panda_set_keywords,                NULL)
	PANDA_METHOD_MAP(setLeading,                panda_set_leading,                 NULL)
	PANDA_METHOD_MAP(setLineCap,                panda_set_line_cap,                NULL)
	PANDA_METHOD_MAP(setLineColor,              panda_set_line_color,              NULL)
	PANDA_METHOD_MAP(setLineDash,               panda_set_line_dash,               NULL)
	PANDA_METHOD_MAP(setLineJoin,               panda_set_line_join,               NULL)
	PANDA_METHOD_MAP(setLineStart,              panda_set_line_start,              NULL)
	PANDA_METHOD_MAP(setLineWidth,              panda_set_line_width,              NULL)
	PANDA_METHOD_MAP(setSubject,                panda_set_subject,                 NULL)
	PANDA_METHOD_MAP(setTextMode,               panda_set_text_mode,               NULL)
	PANDA_METHOD_MAP(setTitle,                  panda_set_title,                   NULL)
	PANDA_METHOD_MAP(setWordSpacing,            panda_set_word_spacing,            NULL)
	PANDA_METHOD_MAP(strokeLine,                panda_stroke_line,                 NULL)
	PANDA_METHOD_MAP(textAnnotation,            panda_text_annotation,             NULL)
	PANDA_METHOD_MAP(textBox,                   panda_text_box,                    NULL)
	PANDA_METHOD_MAP(textDirection,             panda_text_direction,              NULL)
	PANDA_METHOD_MAP(transDuration,             panda_trans_duration,              NULL)
	PANDA_METHOD_MAP(transStyle,                panda_trans_style,                 NULL)
	PANDA_METHOD_MAP(uriLink,                   panda_uri_link,                    NULL)
	{NULL, NULL, NULL}
};
#else
function_entry panda_oo_functions[] = {
	{NULL, NULL, NULL}
};
#endif // PHP_PANDA_OO
/* }}} */

/* {{{ panda_module_entry
 */
zend_module_entry panda_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"panda",
	panda_functions,
	PHP_MINIT(panda),
	PHP_MSHUTDOWN(panda),
	NULL,
	NULL,
	PHP_MINFO(panda),
#if ZEND_MODULE_API_NO >= 20010901
	PHP_PANDA_EXT_VERSION,
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_PANDA
ZEND_GET_MODULE(panda)
#endif

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(panda)
{
	panda_init();

	le_panda = zend_register_list_destructors_ex(destroy_panda_pdf,NULL,le_panda_name,module_number);
	zend_class_entry php_panda_ce;

#ifdef PHP_PANDA_OO
	zend_internal_function fe_set;

	fe_set.type = ZEND_INTERNAL_FUNCTION;
	fe_set.handler = ZEND_FN(panda___set);
	fe_set.function_name = NULL;
	fe_set.scope = NULL;
	fe_set.fn_flags = 0;
	fe_set.prototype = NULL;
	fe_set.num_args = 2;
	fe_set.arg_info = NULL;
	fe_set.pass_rest_by_reference = 0;

	INIT_OVERLOADED_CLASS_ENTRY(php_panda_ce, "panda", panda_oo_functions, NULL, NULL, (zend_function *)&fe_set);
#else
	INIT_CLASS_ENTRY           (php_panda_ce, "panda", panda_oo_functions);
#endif

	php_panda_ce_ptr = zend_register_internal_class(&php_panda_ce TSRMLS_CC);

	REGISTER_STRING_CONSTANT("PANDA_PAGE_SIZE_US_LETTER",			panda_pagesize_usletter, 				CONST_CS | CONST_PERSISTENT);
	REGISTER_STRING_CONSTANT("PANDA_PAGE_SIZE_A4",					panda_pagesize_a4, 						CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_IMAGE_TIFF",						panda_image_tiff, 						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_IMAGE_JPEG",						panda_image_jpeg, 						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_IMAGE_PNG",						panda_image_png, 					    CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_NORMAL",				panda_textmode_normal, 					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_OUTLINE",				panda_textmode_outline, 				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_FILLED_OUTLINE",		panda_textmode_filledoutline,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_INVISIBLE",				panda_textmode_invisible, 				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_FILLED_CLIP",			panda_textmode_filledclipped, 			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_STROKEDCLIPPED",		panda_textmode_strokedclipped, 			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_FILLED_STROKED_CLIPPED",panda_textmode_filledstrokedclipped,	CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_MODE_CLIPPED",				panda_textmode_clipped, 				CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_STROKED",							0,										CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_FILLED",							1, 										CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_QUADRATIC_CURVE_SEGMENT_ONE",		1,										CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_QUADRATIC_CURVE_SEGMENT_TWO",		2,										CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_LINE_CAP_BUTT",					panda_linecap_butt,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_LINE_CAP_ROUND",					panda_linecap_round,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_LINE_CAP_PROJECTED_SQUARE",		panda_linecap_projectedsquare,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_LINE_CAP_MAX",					panda_linecap_max,						CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_LINE_JOIN_MITER",					panda_linejoin_miter,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_LINE_JOIN_ROUND",					panda_linejoin_round,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_LINE_JOIN_BEVEL",					panda_linejoin_bevel,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_LINE_JOIN_MAX",					panda_linejoin_max,						CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_ANNOTATION_FLAG_INVISIBLE",		panda_annotationflag_invisible,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ANNOTATION_FLAG_HIDDEN",			panda_annotationflag_hidden,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ANNOTATION_FLAG_PRINT",			panda_annotationflag_print,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ANNOTATION_FLAG_NO_ZOOM",			panda_annotationflag_nozoom,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ANNOTATION_FLAG_NO_ROTATE",		panda_annotationflag_norotate,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ANNOTATION_FLAG_NO_VIEW",			panda_annotationflag_noview,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ANNOTATION_FLAG_READ_ONLY",		panda_annotationflag_readonly,			CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_ICON_COMMENT",					panda_icon_comment,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ICON_HELP",						panda_icon_help,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ICON_INSERT",						panda_icon_insert,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ICON_KEY",						panda_icon_key,							CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ICON_NEW_PARAGRAPH",				panda_icon_newparagraph,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ICON_NOTE",						panda_icon_note,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_ICON_PARAGRAPH",					panda_icon_paragraph,					CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_SPLIT_YI",				panda_pagetrans_split_yi,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_SPLIT_YO",				panda_pagetrans_split_yo,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_SPLIT_XI",				panda_pagetrans_split_xi,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_SPLIT_XO",				panda_pagetrans_split_xo,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_BLINDS_Y",				panda_pagetrans_blinds_y,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_BLINDS_X",				panda_pagetrans_blinds_x,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_BOX_I",				panda_pagetrans_box_i,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_BOX_O",				panda_pagetrans_box_o,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_WIPE_0",				panda_pagetrans_wipe_0,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_WIPE_90",				panda_pagetrans_wipe_90,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_WIPE_180",				panda_pagetrans_wipe_180,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_WIPE_270",				panda_pagetrans_wipe_270,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_DISSOLVE",				panda_pagetrans_dissolve,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_GLITTER_0",			panda_pagetrans_glitter_0,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_GLITTER_270",			panda_pagetrans_glitter_270,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_GLITTER_315",			panda_pagetrans_glitter_315,			CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_PAGE_TRANS_NONE",					panda_pagetrans_none,					CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_TEXT_DIRECTION_LEFT_TO_RIGHT",	panda_textdirection_l2r,				CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_TEXT_DIRECTION_RIGHT_TO_LEFT",	panda_textdirection_r2l,				CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_HALIGN_LEFT",						panda_halign_left,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_HALIGN_RIGHT",					panda_halign_right,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_HALIGN_CENTER",					panda_halign_center,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_HALIGN_JUSTIFY",					panda_halign_justify,					CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_VALIGN_TOP",						panda_valign_top,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_VALIGN_BOTTOM",					panda_valign_bottom,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_VALIGN_MIDDLE",					panda_valign_middle,					CONST_CS | CONST_PERSISTENT);

	REGISTER_LONG_CONSTANT("PANDA_WINDOW_USE_NONE",					panda_valign_top,						CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_WINDOW_USE_OUTLINES",				panda_valign_bottom,					CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("PANDA_WINDOW_USE_THUMBS",				panda_valign_middle,					CONST_CS | CONST_PERSISTENT);

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(panda)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(panda)
{
	php_info_print_table_start();
	php_info_print_table_row(2, "Panda PDF Library support", "enabled");
#ifdef PHP_PANDA_OO
	php_info_print_table_row(2, "Panda object oriented API", "enabled");
#else
	php_info_print_table_row(2, "Panda object oriented API", "disabled");
#endif
	php_info_print_table_end();
}
/* }}} */

/* {{{ void panda_add_cubic_curve_segment(resource pdf, int x, int y, int cx1, int cy1, int cx2, int cy2)
   Add cubic curve segment. */
PHP_FUNCTION(panda_add_cubic_curve_segment)
{
	zval *pdf_arg;
	long x, y, cx1, cy1, cx2, cy2;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "llllll", &x, &y, &cx1, &cy1, &cx2, &cy2) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ollllll", &pdf_arg, &x, &y, &cx1, &cy1, &cx2, &cy2) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_addcubiccurvesegment(PDF->currPage->page, x, y, cx1, cy1, cx2, cy2);
}
/* }}} */

/* {{{ void panda_add_line_segment(resource pdf, int x, int y)
   Draw to a point. */
PHP_FUNCTION(panda_add_line_segment)
{
	zval *pdf_arg;
	long x, y;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ll", &x, &y) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "oll", &pdf_arg, &x, &y) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_addlinesegment(PDF->currPage->page, x, y);
}
/* }}} */

/* {{{ void panda_add_quadratic_curve_segment(resource pdf, int x, int y, int cx1, int cy1[, int type])
   Add quadratic curve segment. */
PHP_FUNCTION(panda_add_quadratic_curve_segment)
{
	zval *pdf_arg;
	long type, x, y, cx1, cy1;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "llll|l", &x, &y, &cx1, &cy1, &type) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ollll|l", &pdf_arg, &x, &y, &cx1, &cy1, &type) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( ZEND_NUM_ARGS() < 6 ) {
		type = 1;
	}

	switch ( type ) {
		case 1:
			panda_addquadraticcurvesegmentone(PDF->currPage->page, x, y, cx1, cy1);
			break;
		case 2:
			panda_addquadraticcurvesegmenttwo(PDF->currPage->page, x, y, cx1, cy1);
			break;
		default:
			panda_error(panda_false, "Argument 6 invalid.");
	}
}
/* }}} */

/* {{{ void panda_center_window(resource pdf, bool status)
   Ask PDF viewer to center the window when the PDF is loaded. */
PHP_FUNCTION(panda_center_window)
{
	zval *pdf_arg;
	zend_bool status;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "b", &status) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ob", &pdf_arg, &status) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_centerwindow(PDF->PDF, PHP_2_PANDA_BOOL(status));
}
/* }}} */

/* {{{ resource panda_open(string filename)
   Open a new PDF file */
PHP_FUNCTION(panda_open)
{
	char *filename;
	int filename_l;
	php_panda_pdf *PDF;

	if ( zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &filename, &filename_l) == FAILURE)
		WRONG_PARAM_COUNT;

	PDF = emalloc(sizeof(php_panda_pdf));

	PDF->pages = 0;
	PDF->PDF = NULL;
	PDF->currPage = NULL;

	if ( (PDF->PDF = panda_open(filename, "w")) == NULL )
		panda_error(panda_true, "Could not open file.");

	// If leading is not set, line wrapping is defective.
	panda_setleading(PDF->PDF, 16);

	int ret = zend_list_insert(PDF, le_panda);
	object_init_ex((getThis() ? getThis() : return_value), php_panda_ce_ptr);
	add_property_resource((getThis() ? getThis() : return_value), "pdf", ret);
	zend_list_addref(ret);
}
/* }}} */

/* {{{ void panda_close(resource PDF)
   Close a PDF file */
PHP_FUNCTION(panda_close)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_close(PDF->PDF);
	PDF->PDF = NULL;
}
/* }}} */

/* {{{ panda_new_page(resource PDF, string size)
   Creates a new page in PDF */
PHP_FUNCTION(panda_new_page)
{
	zval *pdf_arg;
	char *size;
	int size_l;
	php_panda_pdf *PDF;
	php_panda_page *page, *opage;

	page = opage = NULL;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &size, &size_l) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "os", &pdf_arg, &size, &size_l) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	page = emalloc(sizeof(php_panda_page));

	page->prevPage = NULL;
	page->nextPage = NULL;
	page->page = panda_newpage(PDF->PDF, size);
	if ( PDF->pages > 0 ) {
		opage = PDF->currPage;
		opage->nextPage = page;
		page->prevPage = opage;
	}
	PDF->pages++;
	PDF->currPage = page;
}
/* }}} */

/* {{{ bool panda_prev_page(resource PDF)
   Go to previous page in PDF */
PHP_FUNCTION(panda_prev_page)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( (PDF->pages > 0) && (PDF->currPage->prevPage != NULL) ) {
		PDF->currPage = PDF->currPage->prevPage;
		RETURN_TRUE;
	}
	else {
		RETURN_FALSE;
	}
}
/* }}} */

/* {{{ bool panda_next_page(resource PDF)
   Go to next page in PDF */
PHP_FUNCTION(panda_next_page)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( (PDF->pages > 0) && (PDF->currPage->nextPage != NULL) ) {
		PDF->currPage = PDF->currPage->nextPage;
		RETURN_TRUE;
	}
	else {
		RETURN_FALSE;
	}
}
/* }}} */

/* {{{ void panda_text_box(panda_pdf PDF, string text, int top, int left, int bottom, int right[, double angle[, int hAlign[, int Valign]]])
   Create a text box */
PHP_FUNCTION(panda_text_box)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	long top, left, bottom, right, hAlign, vAlign;
	double angle;
	char *text;
	int text_l, o = (getThis() ? 1 : 0);

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sllll|dll", &text, &text_l, &top, &left, &bottom, &right, &angle, &hAlign, &vAlign) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "osllll|dll", &pdf_arg, &text, &text_l, &top, &left, &bottom, &right, &angle, &hAlign, &vAlign) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( ZEND_NUM_ARGS() == (8 - o) )
		vAlign = panda_valign_top;
	if ( ZEND_NUM_ARGS() == (7 - o) )
		hAlign = panda_halign_left;
	if ( ZEND_NUM_ARGS() == (6 - o) )
		angle = 0.0f;

	panda_textboxrotalign(PDF->PDF, PDF->currPage->page, top, left, bottom, right, angle, hAlign, vAlign, text);
}
/* }}} */

/* {{{ bool panda_set_compression(panda_pdf PDF, int level)
   Set compression on a PDF */
PHP_FUNCTION(panda_set_compression)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	long level;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &level) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &level) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( !(0 < level < 10) ) {
		panda_error(panda_false, "Argument 2 Invalid.");;
	}
	else {
		panda_setproperty(
			PDF->PDF->pages,
			panda_scope_cascade,
			panda_object_property_compress,
			panda_true
		);
		panda_setproperty(
			PDF->PDF->pages,
			panda_scope_cascade,
			panda_object_property_compress_level,
			level
		);
		RETURN_TRUE;
	}
}
/* }}} */

/* {{{ void panda_set_author(panda_pdf PDF, string author)
   Set the author of a PDF. */
PHP_FUNCTION(panda_set_author)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *string;
	int string_l;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &string, &string_l) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "os", &pdf_arg, &string, &string_l) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setauthor(PDF->PDF, string);
}
/* }}} */

/* {{{ void panda_set_font(panda_pdf PDF, char font_name, int type, char encoding)
   Set current font */
PHP_FUNCTION(panda_set_font)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *font_name, *encoding, *font;
	int font_name_l, encoding_l;
	long type;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sls", &font_name, &font_name_l, &type, &encoding, &encoding_l) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "osls", &pdf_arg, &font_name, &font_name_l, &type, &encoding, &encoding_l) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setfont(PDF->PDF, font = panda_createfont(PDF->PDF, font_name, type, encoding));
}
/* }}} */

/* {{{ string panda_make_page_size(double width, double height)
   Return a page size of given width and height suitable for panda_new_page. */
PHP_FUNCTION(panda_make_page_size)
{
	double width, height;
	char retval[32];

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "dd", &width, &height) == FAILURE)
		WRONG_PARAM_COUNT;

	snprintf(retval, sizeof(retval), "[0 0 %.0f %.0f]", width*72, height*72);

	RETURN_STRING(retval, 1);
}
/* }}} */

/* {{{ void panda_image_box(resource pdf, string filename, int top, int left[, double angle[, int bottom[, int right[, int type]]]]) 
   Put an image on a PDF file */
PHP_FUNCTION(panda_image_box)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *filename, *extension;
	int filename_l, x;
	long top, left, bottom, right, type;
	double angle;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sll|dlll", &filename, &filename_l, &top, &left, &angle, &bottom, &right, &type) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "osll|dlll", &pdf_arg, &filename, &filename_l, &top, &left, &angle, &bottom, &right, &type) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	switch ( ZEND_NUM_ARGS() + (getThis() ? 1 : 0) ) {
	case 4:
		angle = 0.0f;
	case 5:
		bottom = 0;
	case 6:
		right = 0;
	case 7:
		extension = strrchr(filename, '.');
		for ( x=0 ; x<strlen(extension) ; x++ )
			extension[x] = tolower(extension[x]);

		if ( extension == NULL ) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Could not determine file type- defaulting to PNG.");
			type = panda_image_png;
		}
		else {
			if ( (strcmp(".jpg", extension) == 0) || (strcmp(".jpeg", extension) == 0 ) || (strcmp(".jpe", extension) == 0) )
				type = panda_image_jpeg;
			else if ( (strcmp(".tif", extension) == 0) || (strcmp(".tiff", extension) == 0 ) )
				type = panda_image_tiff;
			else if ( (strcmp(".png", extension) == 0) )
				type = panda_image_png;
			else {
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Could not determine file type- defaulting to PNG.");
				type = panda_image_png;
			}
		}
		break;
	}

	if ( (bottom == 0) || (right == 0) ) {
		int d_width, d_height;
		panda_imagesize(&d_width, &d_height, filename, type);

		if ( bottom == 0 ) {
            bottom = d_height + top;
		}
		if ( right == 0 ) {
            right = d_width + left;
		}
	}

	panda_imageboxrot(PDF->PDF, PDF->currPage->page, top, left, bottom, right, angle, filename, type);
}
/* }}} */

/* {{{ void panda_set_text_mode(resource pdf, int mode)
   Set font mode. */
PHP_FUNCTION(panda_set_text_mode)
{
	zval *pdf_arg;
	long mode;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &mode) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &mode) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setfontmode(PDF->PDF, mode);
}
/* }}} */

/* {{{ void panda_line_start(resource pdf, int x, int y)
   Set starting coordinates for a line. */
PHP_FUNCTION(panda_set_line_start)
{
	zval *pdf_arg;
	long x, y;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ll", &x, &y) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "oll", &pdf_arg, &x, &y) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setlinestart(PDF->currPage->page, x, y);
}
/* }}} */

/* {{{ void panda_close_line(resource pdf)
   Close line. */
PHP_FUNCTION(panda_close_line)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_closeline(PDF->currPage->page);
}
/* }}} */

/* {{{ void panda_stroke_line(resource pdf)
   Stroke line. */
PHP_FUNCTION(panda_stroke_line)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_strokeline(PDF->currPage->page);
}
/* }}} */

/* {{{ void panda_fill_line(resource pdf)
   Fill line. */
PHP_FUNCTION(panda_fill_line)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_fillline(PDF->currPage->page);
}
/* }}} */

/* {{{ void panda_end_line(resource pdf)
   End line. */
PHP_FUNCTION(panda_end_line)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_endline(PDF->currPage->page);
}
/* }}} */

/* {{{ void panda_set_line_color(resource pdf, int red, int green, int blue)
   Set the color of a line. */
PHP_FUNCTION(panda_set_line_color)
{
	zval *pdf_arg;
	long red, green, blue;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "lll", &red, &green, &blue) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "olll", &pdf_arg, &red, &green, &blue) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( !(0 <= red <= 255) && (0 <= green <= 255) && (0 <= blue <= 255) ) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Arguments #2, #3, and #4 to panda_set_line_color must all be integers between 0 and 255 (inclusive).");
	}
	else {
		panda_setlinecolor(PDF->currPage->page, red, green, blue);
	}
}
/* }}} */

/* {{{ void panda_set_line_width(resource pdf, int width)
   Set the width of a line. */
PHP_FUNCTION(panda_set_line_width)
{
	zval *pdf_arg;
	long width;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &width) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &width) == FAILURE)
		RETURN_FALSE;
	};

	panda_setlinewidth(PDF->currPage->page, width);
}
/* }}} */

/* {{{ void panda_set_line_cap(resource pdf, int cap)
   Set line cap. */
PHP_FUNCTION(panda_set_line_cap)
{
	zval *pdf_arg;
	long cap;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &cap) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &cap) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	switch ( cap ) {
	case panda_linecap_butt:
	case panda_linecap_round:
	case panda_linecap_projectedsquare:
	case panda_linecap_max:
		panda_setlinecap(PDF->currPage->page, cap);
		break;
	default:
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Argument #2 to panda_set_line_cap must be one of either PANDA_LINE_CAP_BUTT, PANDA_LINE_CAP_ROUND, PANDA_LINE_CAP_PROJECTED_SQUARE, or PANDA_LINE_CAP_MAX.");
		break;
	}
}
/* }}} */

/* {{{ void panda_set_line_join(resource pdf, int join)
   Set line joining method. */
PHP_FUNCTION(panda_set_line_join)
{
	zval *pdf_arg;
	long join;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &join) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	switch ( join ) {
	case panda_linejoin_miter:
	case panda_linejoin_round:
	case panda_linejoin_bevel:
	case panda_linejoin_max:
		panda_setlinejoin(PDF->currPage->page, join);
		break;
	default:
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Argument #2 to panda_set_line_join must be one of the following constants: PANDA_LINE_JOIN_MITER, PANDA_LINE_JOIN_ROUND, PANDA_LINE_JOIN_BEVEL, or PANDA_LINE_JOIN_MAX.");
		break;
	}
}
/* }}} */

/* {{{ void panda_set_fill_color(resource pdf, int red, int green, int blue)
   Set fill color. */
PHP_FUNCTION(panda_set_fill_color)
{
	zval *pdf_arg;
	long red, green, blue;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "lll", &red, &green, &blue) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "olll", &pdf_arg, &red, &green, &blue) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( !(0 <= red <= 255) && (0 <= green <= 255) && (0 <= blue <= 255) ) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Arguments #2, #3, and #4 to panda_set_line_color must all be integers between 0 and 255 (inclusive).");
	}
	else {
		panda_setfillcolor(PDF->currPage->page, red, green, blue);
	}
}
/* }}} */

/* {{{ void panda_rectangle(resource pdf, int left, int top, int bottom, int right[, int type])
   Create a rectangle. */
PHP_FUNCTION(panda_rectangle)
{
	zval *pdf_arg;
	long x, y, bottom, right, type;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "llll|l", &x, &y, &bottom, &right, &type) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ollll|l", &pdf_arg, &x, &y, &bottom, &right, &type) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_rectangle (PDF->currPage->page, x, y, bottom, right);

	if ( ZEND_NUM_ARGS() < 6 )
		type = 0;

	switch ( type ) {
	case 0:
		panda_strokeline(PDF->currPage->page);
		break;
	case 1:
		panda_fillline(PDF->currPage->page);
		break;
	default:
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Argument #6 to panda_rectangle must be either PANDA_STROKED or PANDA_FILLED.");
		break;
	}

	panda_endline(PDF->currPage->page);
}
/* }}} */

/* {{{ void panda_make_line(resource pdf, int start_x, int start_y[, int x, int y[, int x, int y[, int x, int y...]]], int end_x, int end_y, int method)
   Build a whole line. Slick. */
PHP_FUNCTION(panda_make_line)
{
	php_panda_pdf *PDF;
	zval ***parr;
	int x, o;
	int arg_count = ZEND_NUM_ARGS();

	parr = emalloc(sizeof(zval) * arg_count);
	o = (getThis() ? 1 : 0);

	if (
			(arg_count < (5 - o)) ||
			(((arg_count) %2) != o ) ||
			(zend_get_parameters_array_ex(arg_count, parr) == FAILURE)
	) {
		WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf((getThis() ? getThis() : *parr[0]) TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setlinestart(PDF->currPage->page, Z_LVAL_P(*parr[1 - o]), Z_LVAL_P(*parr[2 - o]));
	for ( x=3 - o ; x<arg_count-1 ; x+=2 ) {
		panda_addlinesegment(PDF->currPage->page, Z_LVAL_P(*parr[x]), Z_LVAL_P(*parr[x+1]));
	}
	panda_closeline(PDF->currPage->page);
	switch ( Z_LVAL_P(*parr[arg_count-1]) ) {
		case 0:
			panda_strokeline(PDF->currPage->page);
			break;
		case 1:
			panda_fillline(PDF->currPage->page);
			break;
	}
	panda_endline(PDF->currPage->page);

	efree(parr);
}
/* }}} */

/* {{{ void panda_fit_window(resource pdf, bool status)
   Ask viewer to fit the window. */
PHP_FUNCTION(panda_fit_window)
{
	zval *pdf_arg;
	zend_bool status;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "b", &status) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ob", &pdf_arg, &status) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_fitwindow(PDF->PDF, PHP_2_PANDA_BOOL(status));
}
/* }}} */

/* {{{ void panda_full_screen(resource pdf, bool status)
   Ask viewer to render full screen. */
PHP_FUNCTION(panda_full_screen)
{
	zval *pdf_arg;
	zend_bool status;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "b", &status) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ob", &pdf_arg, &status) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_fullscreen(PDF->PDF, PHP_2_PANDA_BOOL(status));
}
/* }}} */

/* {{{ void panda_hide_menu_bar(resource pdf, bool status)
   Ask viewer to hide the menu bar. */
PHP_FUNCTION(panda_hide_menu_bar)
{
	zval *pdf_arg;
	zend_bool status;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "b", &status) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ob", &pdf_arg, &status) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_hidemenubar(PDF->PDF, PHP_2_PANDA_BOOL(status));
}
/* }}} */

/* {{{ void panda_hide_tool_bar(resource pdf, bool status)
   Ask viewer to hide tool bar. */
PHP_FUNCTION(panda_hide_tool_bar)
{
	zval *pdf_arg;
	zend_bool status;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "b", &status) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ob", &pdf_arg, &status) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_hidetoolbar(PDF->PDF, PHP_2_PANDA_BOOL(status));
}
/* }}} */

/* {{{ void panda_hide_window_ui(resource pdf, bool status)
   Ask viewer to hide window UI. */
PHP_FUNCTION(panda_hide_window_ui)
{
	zval *pdf_arg;
	zend_bool status;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "b", &status) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ob", &pdf_arg, &status) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_hidewindowui(PDF->PDF, PHP_2_PANDA_BOOL(status));
}
/* }}} */

/* {{{ void panda_nfs_page_mode(resource pdf, int mode)
   Ask viewer to set NFS page mode. */
PHP_FUNCTION(panda_nfs_page_mode)
{
	zval *pdf_arg;
	long type;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &type) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	switch ( type ) {
	case panda_window_usenone:
	case panda_window_useoutlines:
	case panda_window_usethumbs:
		panda_nfspagemode(PDF->PDF, type);
		break;
	default:
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Argument #2 to panda_nfs_page_mode must be either PANDA_WINDOW_USE_NONE, PANDA_WINDOW_USE_OUTLINES, or PANDA_WINDOW_USE_THUMBS.");
		break;
	}
}
/* }}} */

/* {{{ void panda_page_duration(resource pdf, int seconds)
   Ask viewer to move on the the next page in int seconds. */
PHP_FUNCTION(panda_page_duration)
{
	zval *pdf_arg;
	long duration;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &duration) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &duration) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_pageduration(PDF->PDF, PDF->currPage->page, duration);
}
/* }}} */

/* {{{ void panda_set_font_size(resource pdf, int size [, int leading])
   Set the font size. */
PHP_FUNCTION(panda_set_font_size)
{
	zval *pdf_arg;
	long size, lead;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l|l", &size, &lead) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol|l", &pdf_arg, &size, &lead) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setleading(PDF->PDF, (ZEND_NUM_ARGS() < 3) ? (size * 1.2) : lead);
	panda_setfontsize(PDF->PDF, size);
}
/* }}} */

/* {{{ void panda_set_horizontal_scaling(resource pdf, double amount)
   Set the horizontal scaling. */
PHP_FUNCTION(panda_set_horizontal_scaling)
{
	zval *pdf_arg;
	double amount;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "d", &amount) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "od", &pdf_arg, &amount) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_sethorizontalscaling(PDF->PDF, amount);
}
/* }}} */

/* {{{ void panda_set_keywords(resource pdf, string keywords)
   Set document keywords. */
PHP_FUNCTION(panda_set_keywords)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *keywords;
	int keywords_l;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &keywords, &keywords_l) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "os", &pdf_arg, &keywords, &keywords_l) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setkeywords(PDF->PDF, keywords);
}
/* }}} */

/* {{{ void panda_set_leading(resource pdf, double amount)
   Set the leading. */
PHP_FUNCTION(panda_set_leading)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	double amount;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "d", &amount) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "od", &pdf_arg, &amount) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setleading(PDF->PDF, amount);
}
/* }}} */

/* {{{ void panda_set_line_dash(resource pdf, int on, int off, int phase)
   Set line dashing. */
PHP_FUNCTION(panda_set_line_dash)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	long on, off, phase;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "lll", &on, &off, &phase) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "olll", &pdf_arg, &on, &off, &phase) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setlinedash(PDF->currPage->page, on, off, phase);
}
/* }}} */

/* {{{ void panda_set_subject(resource pdf, string subject)
   Set PDF subject. */
PHP_FUNCTION(panda_set_subject)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *subject;
	int subject_l;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &subject, &subject_l) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "os", &pdf_arg, &subject, &subject_l) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setsubject(PDF->PDF, subject);
}
/* }}} */

/* {{{ void panda_set_title(resource pdf, string title)
   Set the title of a PDF document. */
PHP_FUNCTION(panda_set_title)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *title;
	int title_l;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &title, &title_l) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "os", &pdf_arg, &title, &title_l) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_settitle(PDF->PDF, title);
}
/* }}} */

/* {{{ void panda_set_word_spacing(resource pdf, double amount)
   Set word spacing. */
PHP_FUNCTION(panda_set_word_spacing)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	double amount;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "d", &amount) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "od", &pdf_arg, &amount) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_setwordspacing(PDF->PDF, amount);
}
/* }}} */

/* {{{ void panda_text_annotation(resource pdf, string text, string title, int top, int left, int bottom, int right, double red, double green, double blue, int flags[, int open, int icon])
   Create a text annotation. */
PHP_FUNCTION(panda_text_annotation)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *text, *title;
	int text_l, title_l;
	long top, left, bottom, right, flags, open, icon;
	double red, green, blue;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sslllldddl|ll", &text, &text_l, &title, &title_l, &top, &left, &bottom, &right, &red, &green, &blue, &flags, &open, &icon) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "osslllldddl|ll", &pdf_arg, &text, &text_l, &title, &title_l, &top, &left, &bottom, &right, &red, &green, &blue, &flags, &open, &icon) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	switch ( ZEND_NUM_ARGS() ) {
		case 13:
			panda_textannotation(PDF->PDF, PDF->currPage->page, open, text, title, top, left, bottom, right, red, green, blue, icon, flags);
			break;
		case 11:
			panda_freetextannotation(PDF->PDF, PDF->currPage->page, text, title, top, left, bottom, right, red, green, blue, flags);
			break;
		default:
			printf("%d\n", ZEND_NUM_ARGS());
			WRONG_PARAM_COUNT;
	}
}
/* }}} */

/* {{{ void panda_trans_duration(resource pdf, double seconds)
   Set transition duration. */
PHP_FUNCTION(panda_trans_duration)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	double seconds;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "d", &seconds) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "od", &pdf_arg, &seconds) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( seconds < 0 ) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Argument #2 to panda_trans_duration must be positive.");
		seconds = 0;
	}
	panda_transduration(PDF->PDF, PDF->currPage->page, seconds);
}
/* }}} */

/* {{{ void panda_trans_style(resource pdf, int style)
   Set transition style. */
PHP_FUNCTION(panda_trans_style)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	long style;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &style) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &style) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if (
		(style == panda_pagetrans_split_yi) ||
		(style == panda_pagetrans_split_yo) ||
		(style == panda_pagetrans_split_xi) ||
		(style == panda_pagetrans_split_xo) ||
		(style == panda_pagetrans_blinds_y) ||
		(style == panda_pagetrans_blinds_x) ||
		(style == panda_pagetrans_box_i) ||
		(style == panda_pagetrans_box_o) ||
		(style == panda_pagetrans_wipe_0) ||
		(style == panda_pagetrans_wipe_90) ||
		(style == panda_pagetrans_wipe_180) ||
		(style == panda_pagetrans_wipe_270) ||
		(style == panda_pagetrans_dissolve) ||
		(style == panda_pagetrans_glitter_0) ||
		(style == panda_pagetrans_glitter_270) ||
		(style == panda_pagetrans_glitter_315) ||
		(style == panda_pagetrans_none)
	) {
		panda_transstyle(PDF->PDF, PDF->currPage->page, style);
	}
}
/* }}} */

/* {{{ void panda_text_direction(resource pdf, int direction)
   Change the direction the text goes. */
PHP_FUNCTION(panda_text_direction)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	long direction;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &direction) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "ol", &pdf_arg, &direction) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	if ( (direction == panda_textdirection_l2r) || (direction == panda_textdirection_r2l) )
		panda_textdirection(PDF->PDF, direction);
	else
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Argument #2 to panda_text_direction must be either PANDA_TEXT_DIRECTION_LEFT_TO_RIGHT or PANDA_TEXT_DIRECTION_RIGHT_TO_LEFT");
}
/* }}} */

/* {{{ void panda_rewind(resource pdf)
   Rewind resource pdf to first page. */
PHP_FUNCTION(panda_rewind)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	php_panda_pdf_rw(PDF);
}
/* }}} */

/* {{{ void panda_fast_forward(resource pdf)
   Fast forward resource pdf to last page. */
PHP_FUNCTION(panda_fast_forward)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	php_panda_pdf_ff(PDF);
}
/* }}} */

/* {{{ int panda_get_num_pages(resource pdf)
   End line. */
PHP_FUNCTION(panda_get_num_pages)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	RETURN_LONG(PDF->pages);
}
/* }}} */

/* {{{ double panda_get_string_width(resource pdf, string string)
   Get width of string with current settings (font style, size, etc.) */
PHP_FUNCTION(panda_get_string_width)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *string;
	int string_l;
	double width;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &string, &string_l) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "os", &pdf_arg, &string, &string_l) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	width = panda_stringwidth(PDF->PDF, string, NULL);

	RETURN_DOUBLE(width);
}
/* }}} */

/* {{{ int panda_find_line_break(resource pdf, string string, int wrapWidth, double distance)
   Number of characters before wrapping occurs. */
PHP_FUNCTION(panda_find_line_break)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *string;
	int string_l, position;
	long wrapWidth;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sl", &string, &string_l, &wrapWidth) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "osl", &pdf_arg, &string, &string_l, &wrapWidth) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	position = panda_findlinebreak(PDF->PDF, string, panda_getfontmetric(PDF->PDF), wrapWidth, NULL);

	RETURN_LONG(PDF->pages);
}
/* }}} */

/* {{{ int panda_get_page_width(resource pdf)
   Get the width of the current page. */
PHP_FUNCTION(panda_get_page_width)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	RETURN_LONG(PDF->currPage->page->width);
}
/* }}} */

/* {{{ int panda_get_page_height(resource pdf)
   Get the height of the current page. */
PHP_FUNCTION(panda_get_page_height)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;

	if ( getThis() ) {
		if ( ZEND_NUM_ARGS() != 0 )
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "o", &pdf_arg) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	RETURN_LONG(PDF->currPage->page->height);
}
/* }}} */

/* {{{ void panda_uri_link(resource pdf, string uri, int top, int left, int bottom, int right[, int border])
   Get current page size. */
PHP_FUNCTION(panda_uri_link)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	char *uri;
	int uri_l;
	long top, left, bottom, right, border = 0;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sllll|l", &uri, &uri_l, &top, &left, &bottom, &right, &border) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "osllll|l", &pdf_arg, &uri, &uri_l, &top, &left, &bottom, &right, &border) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	panda_urilink(PDF->PDF, PDF->currPage->page, uri, top, left, bottom, right, border);
}
/* }}} */

/* {{{ void panda_link(resource pdf, int dest_page, int dest_top, int dest_left, int top, int left, int bottom, int right[, int border[, int zoom]])
   Get current page size. */
PHP_FUNCTION(panda_link)
{
	zval *pdf_arg;
	php_panda_pdf *PDF;
	php_panda_page *tPage;
	int x;
	long dest_page, dest_top, dest_left, zoom, top, left, bottom, right, border = 0;

	if ( getThis() ) {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "lllllll|ll", &dest_page, &dest_top, &dest_left, &top, &left, &bottom, &right, &border, &zoom) == FAILURE)
			WRONG_PARAM_COUNT;
		pdf_arg = getThis();
	}
	else {
		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "olllllll|ll", &pdf_arg, &dest_page, &dest_top, &dest_left, &top, &left, &bottom, &right, &border, &zoom) == FAILURE)
			WRONG_PARAM_COUNT;
	}
	if ( (PDF = php_panda_get_pdf(pdf_arg TSRMLS_CC)) == NULL ) {
		RETURN_FALSE;
	};

	tPage = PDF->currPage;
	php_panda_pdf_rw(PDF);

	for ( x = 0 ; x < dest_page-1 ; x++ )
		if ( PDF->currPage->nextPage != NULL )
			PDF->currPage = PDF->currPage->nextPage;

	panda_link(PDF->PDF, PDF->currPage->page, dest_top, dest_left, zoom, tPage->page, top, left, bottom, right, border);

	PDF->currPage = tPage;
}
/* }}} */

#ifdef PHP_PANDA_OO
/* {{{ void panda::__set(string property, mixed value)
   Overloaded properties */
PHP_METHOD(panda, __set)
{
    zval *val;
    char *var;
    int var_l;
	php_panda_pdf *PDF;

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz", &var, &var_l, &val) == FAILURE)
        WRONG_PARAM_COUNT;

    if ( strcasecmp("pdf", var) != 0 ) {
        if ( (PDF = php_panda_get_pdf(getThis() TSRMLS_CC)) == NULL ) {
            RETURN_FALSE;
        }

        if ( strcasecmp(var, "font") == 0 ) {
            convert_to_string(val);
            panda_setfont(PDF->PDF, Z_STRVAL_P(val));
        }
        else if ( strcasecmp(var, "fontSize") == 0 ) {
            convert_to_long(val);
            panda_setfontsize(PDF->PDF, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "fontMode") == 0 ) {
            convert_to_long(val);
            panda_setfontmode(PDF->PDF, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "characterSpacing") == 0 ) {
            convert_to_double(val);
            panda_setcharacterspacing(PDF->PDF, Z_DVAL_P(val));
        }
        else if ( strcasecmp(var, "wordSpacing") == 0 ) {
            convert_to_double(val);
            panda_setwordspacing(PDF->PDF, Z_DVAL_P(val));
        }
        else if ( strcasecmp(var, "horizontalScaling") == 0 ) {
            convert_to_double(val);
            panda_sethorizontalscaling(PDF->PDF, Z_DVAL_P(val));
        }
        else if ( strcasecmp(var, "leading") == 0 ) {
            convert_to_double(val);
            panda_setleading(PDF->PDF, Z_DVAL_P(val));
        }
        else if ( strcasecmp(var, "lineWidth") == 0 ) {
            convert_to_long(val);
            panda_setlinewidth(PDF->currPage->page, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "lineCap") == 0 ) {
            convert_to_long(val);
            panda_setlinecap(PDF->currPage->page, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "lineJoin") == 0 ) {
            convert_to_long(val);
            panda_setlinejoin(PDF->currPage->page, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "author") == 0 ) {
            convert_to_string(val);
            panda_setauthor(PDF->PDF, Z_STRVAL_P(val));
        }
        else if ( strcasecmp(var, "title") == 0 ) {
            convert_to_string(val);
            panda_settitle(PDF->PDF, Z_STRVAL_P(val));
        }
        else if ( strcasecmp(var, "subject") == 0 ) {
            convert_to_string(val);
            panda_setsubject(PDF->PDF, Z_STRVAL_P(val));
        }
        else if ( strcasecmp(var, "keywords") == 0 ) {
            convert_to_string(val);
            panda_setkeywords(PDF->PDF, Z_STRVAL_P(val));
        }
        else if ( strcasecmp(var, "hideToolbar") == 0 ) {
            convert_to_long(val);
            panda_hidetoolbar(PDF->PDF, PHP_2_PANDA_BOOL(Z_LVAL_P(val)));
        }
        else if ( strcasecmp(var, "hideMenubar") == 0 ) {
            convert_to_long(val);
            panda_hidemenubar(PDF->PDF, PHP_2_PANDA_BOOL(Z_LVAL_P(val)));
        }
        else if ( strcasecmp(var, "hideWindowUI") == 0 ) {
            convert_to_long(val);
            panda_hidewindowui(PDF->PDF, PHP_2_PANDA_BOOL(Z_LVAL_P(val)));
        }
        else if ( strcasecmp(var, "nfsPageMode") == 0 ) {
            convert_to_long(val);
            panda_nfspagemode(PDF->PDF, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "fullscreen") == 0 ) {
            convert_to_long(val);
            panda_fullscreen(PDF->PDF, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "textDirection") == 0 ) {
            convert_to_long(val);
            panda_textdirection(PDF->PDF, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "pageDuration") == 0 ) {
            convert_to_long(val);
            panda_pageduration(PDF->PDF, PDF->currPage->page, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "transDuration") == 0 ) {
            convert_to_long(val);
            panda_transduration(PDF->PDF, PDF->currPage->page, Z_LVAL_P(val));
        }
        else if ( strcasecmp(var, "transStyle") == 0 ) {
            convert_to_long(val);
            panda_transstyle(PDF->PDF, PDF->currPage->page, Z_LVAL_P(val));
        }
    }

    zval *tmp;
    if ( zend_hash_find(Z_OBJPROP_P(getThis()), var, var_l, (void *)&tmp) == SUCCESS ) { // Property exists
        zend_hash_index_update(Z_OBJPROP_P(getThis()), Z_LVAL_P(tmp), (void *)&val, sizeof(zval *), NULL);
    }
    else { // Property doesn't exist
        add_property_zval(getThis(), var, val);
    }
}
/* }}} */
#endif // PHP_PANDA_OO

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
